using System;

namespace BusinessTier
{

	/// <summary>
	/// This class represents a single bank customer.
	/// </summary>
	public class BankCustomer : ICloneable, IComparable
	{
		private int      m_CustomerID;
		private string   m_FirstName; 
		private string   m_LastName;
		private decimal  m_Balance;

		/// <summary>
		/// Constructs a new BankCustomer from given id, name and balance.
		/// </summary>
		/// <param name="cid"></param>
		/// <param name="fn"></param>
		/// <param name="ln"></param>
		/// <param name="balance"></param>
		public BankCustomer(int cid, string fn, string ln, decimal balance)
		{
			this.m_CustomerID = cid;
			this.m_FirstName  = fn;
			this.m_LastName   = ln;
			this.m_Balance    = balance;
		}

		/// <summary>
		/// Constructs a new BankCustomer from given name; balance is set to 0.00.
		/// </summary>
		/// <param name="cid"></param>
		/// <param name="fn"></param>
		/// <param name="ln"></param>
		public BankCustomer(int cid, string fn, string ln)  // initial balance = 0.00
		: this(cid, fn, ln, 0.0M)
		{ }

		/// <summary>
		/// Allows access to customer's ID.
		/// </summary>
		public int CustomerID
		{
			get { return this.m_CustomerID; }
		}

		/// <summary>
		/// Allows access to customer's first name.
		/// </summary>
		public string FirstName
		{
			get { return this.m_FirstName; }
		}

		/// <summary>
		/// Allows access to customer's last name.
		/// </summary>
		public string LastName
		{
			get { return this.m_LastName; }
		}

		/// <summary>
		/// Allows access to customer's raw balance.
		/// </summary>
		public decimal Balance
		{
			get { return this.m_Balance; }
		}

		/// <summary>
		/// Provides balance as a nicely-formatted string.
		/// </summary>
		public string FormattedBalance
		{
			get { return this.GetFormattedBalance(); }
		}

		/// <summary>
		/// Returns balance formatted according to control panel's locality setting.
		/// </summary>
		/// <returns>currency-formatted string (e.g. "$1,000.99")</returns>
		public string GetFormattedBalance()
		{
			return this.Balance.ToString("C", System.Globalization.NumberFormatInfo.CurrentInfo);
		}

		//
		// methods
		//

		/// <summary>
		/// Deposits the given amount into the customer's account.
		/// </summary>
		/// <param name="amt">amount to deposit (must be >= 0)</param>
		public void Deposit(decimal amt)
		{
			if (amt < 0)
				throw new System.ArgumentException("BankCustomer.Deposit(amt): negative amount.");
			else
				this.m_Balance += amt;
		}

		/// <summary>
		/// Withdraws the given amount from the customers account.
		/// </summary>
		/// <param name="amt">amount to withdraw (must be >= 0)</param>
		/// <exception cref="Exception">Throw if sufficient funds are not available</exception>
		public void Withdraw(decimal amt)
		{
			if (amt < 0)  // negative!
				throw new System.ArgumentException("BankCustomer.Withdraw(amt): negative amount.");
			else if (amt > this.Balance)  // insufficient funds!
				throw new System.ApplicationException("BankCustomer.Withdraw(amt): insufficient funds.");
			else  // go ahead, withdraw...
				this.m_Balance -= amt;
		}

		/// <summary>
		/// Returns string-based representation of object.
		/// </summary>
		/// <returns>string in format "lastname, firstname"</returns>
		public override string ToString()
		{
			return this.LastName + ", " + this.FirstName;
		}

		/// <summary>
		/// Returns true if THIS object has the same last and first name as the given object obj; comparison is
		/// case-sensitive.
		/// </summary>
		/// <param name="obj">object to be compared to (must be an instance of BankCustomer)</param>
		/// <returns>true if names are equal, false if not (case-sensitive compare)</returns>
		public override bool Equals(object obj)
		{
			if (obj == null) 
				return false;
			if ((obj.GetType().Equals(this.GetType())) == false)
				return false;

			BankCustomer  other;
			other = (BankCustomer) obj;

			return this.FirstName.Equals(other.FirstName) && 
				   this.LastName.Equals(other.LastName);
		}

		/// <summary>
		/// Returns an integer hash code for this object.
		/// </summary>
		/// <returns>integer code appropraite for hashing</returns>
		public override int GetHashCode()
		{
			return this.FirstName.GetHashCode() + this.LastName.GetHashCode();
		}

		/// <summary>
		/// Returns a shallow copy of this object.
		/// </summary>
		/// <returns>shallow copy</returns>
		public object Clone()
		{
			return this.MemberwiseClone();  // shallow copy
		}

		/// <summary>
		/// Compares customers based on their customer id.
		/// </summary>
		/// <param name="obj"></param>
		/// <returns>-1 if this customer is < obj, 0 if ==, 1 if ></returns>
		public int CompareTo(Object obj)
		{
			// sort customers based on their ID numbers  
			return this.CustomerID.CompareTo( ((BankCustomer)obj).CustomerID );
		}

		/// <summary>
		/// Comparator class for sorting customers by last name, first name.
		/// </summary>
		public class SortByLastName : System.Collections.IComparer
		{
			public int Compare(object o1, object o2)
			{
				BankCustomer c1 = (BankCustomer) o1;
				BankCustomer c2 = (BankCustomer) o2;

				// compare by last name, then first name...
				int  diff = c1.LastName.CompareTo(c2.LastName);
				if (diff == 0) diff = c1.FirstName.CompareTo(c2.FirstName);

				return diff;
			}
		}

		/// <summary>
		/// Comparator class for sorting customers by balance, in descending order.
		/// </summary>
		public class SortByBalance : System.Collections.IComparer
		{
			public int Compare(object o1, object o2)
			{
				BankCustomer c1 = (BankCustomer) o1;
				BankCustomer c2 = (BankCustomer) o2;

				// compare by account balance, yielding descending order...
				return -c1.Balance.CompareTo(c2.Balance);
			}
		}

	}//class
}//namespace
